#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <pwd.h>
#include <sys/types.h>
#include <unistd.h>
#include <signal.h>
#include <sys/wait.h>
#include <sys/stat.h>
#include <fcntl.h>

/* Catalina's options.  We validate these in order to a) print
 * sensible usage messages, and b) avoid usage errors from catalina
 * which may end up in logfiles.
 */
static const struct {
    char *name;
    int has_arg;
} catalina_opts[] = {{"-config", 1}, {"-debug"}, {"-nonaming"}, {NULL}};

static void error_opt_requires_arg(const char *opt)
{
    fprintf(stderr, "tomcat: option requires an argument -- `%s'\n", opt);
    fprintf(stderr, "Try `tomcat --help' for more information.\n");
    exit(EXIT_FAILURE);
}
static void error_unknown_opt(const char *opt)
{
    fprintf(stderr, "tomcat: unrecognized option -- `%s'\n", opt);
    fprintf(stderr, "Try `tomcat --help' for more information.\n");
    exit(EXIT_FAILURE);
}
static void usage(void)
{
    printf("Usage: tomcat [OPTION]\n\n");
    printf("  -config FILE      use this alternative configuration file\n");
    printf("  -DVAR=VAL         define system property VAR with value VAL\n");
    printf("  -help             display this help and exit\n");
    printf("  -home DIR         run catalina from within DIR\n");
    printf("  -logfile FILE     log messages to this file\n");
    /* printf("  -lowports         ...\n"); */
    printf("  -nonaming         don't use JNDI\n");
    printf("  -pidfile FILE     file to write the process ID to\n");
    /* printf("  -security         ...\n"); */
    printf("  -user UID         run the daemon as user UID\n");
    printf("\nFor bug reporting instructions please see " \
	"http://bugzilla.redhat.com/\n");
    exit(EXIT_SUCCESS);
}

/* Distinguish failures in this file from failures in the rest of the
 * program (not that catalina is particularly consistent with regard
 * to exit codes).  And yes, I know about the two EXIT_FAILURES above.
 */
#ifdef EXIT_FAILURE
#undef EXIT_FAILURE
#endif
#define EXIT_FAILURE 234

static void perror_exit(const char *prefix)
{
    perror(prefix);
    exit(EXIT_FAILURE);
}

/* A file descriptor upon which the initial process waits to be told
 * to exit and thus detach us from our controlling tty.  A SIGCHLD
 * handler ensures sensible behaviour in the event of the child's
 * premature death.
 */
static int write_fd = -1;

#define CONTROL_CHILD_DIED 1
#define CONTROL_BACKGROUND 2

static int send_command(char cmd)
{
    ssize_t ret = write(write_fd, &cmd, 1);
    if (ret == -1)
	perror("tomcat: write");
    return ret == 1 ? 0 : -1;
}
int daemon_init(void)
{
    return send_command(CONTROL_BACKGROUND);
}
static void sigchld_handler(int sig)
{
    send_command(CONTROL_CHILD_DIED);
}

/* Allow the child to delete the pidfile on exiting.
 */
static const char *pidfile = NULL;
void daemon_cleanup(void)
{
    if (pidfile)
	unlink(pidfile);
}

/* We export the logfile redirection functionality so that the child
 * can restart the logs on demand.  This avoids the need for a
 * complete restart for logrotate.
 */
static const char *logfile = NULL;
static void redirect_stdout_stderr(void)
{
    int fd;

    fd = open(logfile, O_WRONLY|O_CREAT|O_APPEND,
	      S_IRUSR|S_IWUSR|S_IRGRP|S_IROTH);
    if (fd == -1)
	perror_exit("tomcat: open");

    fflush(stdout);
    if (dup2(fd, 1) != 1)
	perror_exit("tomcat: dup2");

    fflush(stderr);
    if (dup2(fd, 2) != 2)
	perror_exit("tomcat: dup2");
}
void daemon_recreate_log(void)
{
    redirect_stdout_stderr();
}

/* Link with -Wl,--wrap,main so that __wrap_main() is called in the
 * real main()'s stead.  Or compile with -DTEST to enable the testing
 * code at the bottom of the file.
 */
int __wrap_main(int argc, const char **argv)
{
    const char *home = NULL;
    const char *user = NULL;
    /* int lowports = 0; */
    /* int security = 0; */
    
    /* Extract our options, rearranging argv as we go.  We also
     * validate the options that we are going to pass through so
     * that a sensible usage message may be printed and to avoid
     * leaving usage messages in logfiles.
     */
    {
	int orig_argc = argc;
	int i;
	
	for (i = 1; i < argc; i++) {
	    const char **argopt = NULL;
	    int eat_argv = 0;

	    if (!strcmp(argv[i], "-help"))
		usage();
	    else if (!strcmp(argv[i], "-home"))
		argopt = &home;
	     else if (!strcmp(argv[i], "-logfile"))
		argopt = &logfile;
	    else if (!strcmp(argv[i], "-pidfile"))
		argopt = &pidfile;
	    else if (!strcmp(argv[i], "-user"))
		argopt = &user;
	    /* else if (!strcmp(argv[i], "-lowports")) {
		lowports = 1;
		eat_argv = 1;
	    } */
	    /* else if (!strcmp(argv[i], "-security")) {
		security = 1;
		eat_argv = 1;
	    } */
	    else if (!strncmp(argv[i], "-D", 2)) {
		if (strlen(argv[i]) <= 2)
		    error_opt_requires_arg(argv[i]);
	    }
	    else {
		int j;
		
		for (j = 0; catalina_opts[j].name; j++) {
		    if (!strcmp(argv[i], catalina_opts[j].name)) {
			if (catalina_opts[j].has_arg) {
			    if (i >= argc - 1)
				error_opt_requires_arg(argv[i]);
			    i++;
			}
			break;
		    }
		}
		if (!catalina_opts[j].name)
		    error_unknown_opt(argv[i]);
	    }

	    if (argopt) {
		if (i >= argc - 1)
		    error_opt_requires_arg(argv[i]);
		*argopt = argv[i + 1];
		eat_argv = 2;
	    }

	    if (eat_argv) {
		int j;
		
		argc -= eat_argv;
		for (j = i; j < argc; j++)
		    argv[j] = argv[j + eat_argv];
		i--;
	    }
	}

	/* Add the stupid 'start' option that catalina is expecting */
	if (argc + 1 < orig_argc) {
	    /* XXX can we just realloc() argv? */
	    const char **new_argv = calloc(argc + 1, sizeof(char *));
	    if (new_argv == NULL)
		perror_exit("tomcat: calloc");
	    for (i = 0; i < argc; i++)
		new_argv[i] = argv[i];
	    argv = new_argv;
	}
	argv[argc++] = "start";
    }
    
    /* First of all lets lose those privileges */
    if (user) {
	struct passwd *entry;

	errno = 0;
	entry = getpwnam(user);
	if (entry == NULL) {
	    if (errno)
		perror("tomcat: getpwnam");
	    else
		fprintf(stderr, "tomcat: unknown user `%s'\n", user);
	    exit(EXIT_FAILURE);
	}

	if (setgid(entry->pw_gid) != 0)
	    perror_exit("tomcat: setgid");

	if (setuid(entry->pw_uid) != 0)
	    perror_exit("tomcat: setuid");
    }
    
    /* Change directory to the required location
     */
    if (home) {
	if (chdir(home) != 0)
	    perror_exit("tomcat: chdir");
    }

    /* Make sure there are no stale pidfiles lying around
     */
    if (pidfile) {
	if (unlink(pidfile) == 0)
	    fprintf(stderr, "tomcat: warning: deleted old pidfile\n");
    }
    
    /* Fork and, if the parent, wait to be told to exit by the child
     * (or alternativly for the child to die).  We set and the reset
     * the SIGCHLD handler to avoid a potential race which could cause
     * the parent to hang.
     */
    {
	int read_fd;
	sig_t orig_handler;
	pid_t pid;

	{
	    int fds[2];
	    if (pipe(fds) != 0)
		perror_exit("tomcat: pipe");
	    read_fd = fds[0];
	    write_fd = fds[1];
	}

	orig_handler = signal(SIGCHLD, sigchld_handler);
	if (orig_handler == SIG_ERR)
	    perror_exit("tomcat: signal");

	pid = fork();
	if (pid == -1)
	    perror_exit("tomcat: fork");

	if (pid == 0) {
	    if (signal(SIGCHLD, orig_handler) == SIG_ERR)
		perror_exit("tomcat: signal");
	    close(read_fd);
	}
	else {
	    char cmd;
	    while (1) {
		int ret = read(read_fd, &cmd, 1);
		if (ret == 1)
		    break;
		else if (ret == -1 && errno != EINTR)
		    perror_exit("tomcat: read");
	    }
	    if (cmd == CONTROL_BACKGROUND) {
		
		/* The child has successfully initialized so write the
		 * pidfile and exit, thus backgrounding ourself.
		 */
		if (pidfile) {
		    FILE *fp = fopen(pidfile, "w");
		    if (fp) {
			fprintf(fp, "%d\n", pid);
			fclose(fp);
		    }
		    else
			perror("tomcat: fopen");
		}

		exit(EXIT_SUCCESS);
	    }
	    else if (cmd == CONTROL_CHILD_DIED) {

		/* The child died during initialisation.
		 */
		int status;
		if (waitpid(pid, &status, 0) == -1)
		    perror_exit("tomcat: waitpid");
		if (WIFEXITED(status)) {
		    status = WEXITSTATUS(status);
		    /* Catalina is, ahem, less than consistent here */
		    if (status == EXIT_SUCCESS)
			status = EXIT_FAILURE;
		    exit(status);
		}
		else if (WIFSIGNALED(status))
		    fprintf(stderr, "tomcat: killed with signal %d\n",
			    WTERMSIG(status));
		else
		    fprintf(stderr, "tomcat: internal error 1 (%d)\n", status);
	    }
	    else	
		fprintf(stderr, "tomcat: internal error 2 (%d)\n", cmd);

	    exit(EXIT_FAILURE);
	}
    }

    /* Redirect stdout and stderr to the specified logfile
     */
    if (logfile)
	redirect_stdout_stderr();
    
    /* Finally we are ready to jump into Java land and run catalina.
     */
    return __real_main(argc, argv);
}

#ifdef TEST
int main(int argc, const char **argv)
{
    return __wrap_main(argc, argv);
}

int __real_main(int argc, const char **argv)
{
    return 42;
}
#endif /* TEST */
